/* shell variable functions:
 *	Used to load or retrieve shell variable information from the
 *	shell variable table.  Depending on how the monitor is built,
 *	this can be used a few different ways...
 *
 *	1. Build with INCLUDE_FLASHENV on:
 *	This assumes TFS is not included, but still provides a primitive
 *	means of storing some variables in non-volatile flash so that at
 *	bootup, a basic environment can be initialized.
 *
 *	2. Build with INCLUDE_FLASHENV off:
 *	This assumes TFS is included, so the environment is initialized
 *	when monrc is run.
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Note1: the majority of this code was edited with 4-space tabs.
 *	Note2: as more and more contributions are accepted, the term "author"
 *		   is becoming a mis-representation of credit.
 *
 *	Original author:	Ed Sutter
 *	Email:				ed.sutter@alcatel-lucent.com
 *	Phone:				908-582-2351
 */
#include <stdarg.h>
#include "config.h"
#include "tfs.h"
#include "tfsprivate.h"
#include "ether.h"
#include "genlib.h"
#include "stddefs.h"
#include "cli.h"
#include "version.h"
#include "boardinfo.h"
#include "ee.h"
#include "timer.h"
#include "flashenv.h"

char *whatplatform = "@(#)PLATFORM=" PLATFORM_NAME;

static int shell_print(void);
static void shell_clear(void);
static int shell_varcheck(int);

/* Structure used for the shell variables: */
struct s_shell {
	char	*val;		/* Value stored in shell variable */
	char	*name;		/* Name of shell variable */
	int		vsize;		/* Size of storage allocated for value */
	struct	s_shell *next;
};

struct s_shell *shell_vars;

/* If no malloc, then use locally defined env_alloc() and env_free()...
 */
#if INCLUDE_MALLOC

#define env_alloc	malloc
#define env_free	free

#else

#ifndef ENV_ALLOC_TOT
#define ENV_ALLOC_TOT	24
#endif
#define ENV_ALLOC_SIZE	(sizeof(struct s_shell)+32)

struct env_space {
	int inuse;
	char space[ENV_ALLOC_SIZE];
} envSpace[ENV_ALLOC_TOT];


char *
env_alloc(int size)
{
	int	 i;

	if (size >= ENV_ALLOC_SIZE)
		return(0);

	for(i=0;i<ENV_ALLOC_TOT;i++) {
		if (envSpace[i].inuse == 0) {
			envSpace[i].inuse = 1;
			memset(envSpace[i].space,0,ENV_ALLOC_SIZE);
			return(envSpace[i].space);
		}
	}
	return(0);
}

void
env_free(char *space)
{
	int i;

	for(i=0;i<ENV_ALLOC_TOT;i++) {
		if (envSpace[i].space == space) {
			envSpace[i].inuse = 0;
			break;
		}
	}
	return;
}
#endif

/*
 *  Set()
 *
 *	Syntax:
 *		set var				clears the variable 'var'
 *		set var value		assign "value" to variable 'var'
 */
char *SetHelp[] = {
	"Shell var ops",
#if INCLUDE_FLASHENV
	"-[cf] [var] [val]",
#else
	"-[c] [var] [val]",
#endif
#if INCLUDE_VERBOSEHELP
	" -c        clear the environment",
#if INCLUDE_FLASHENV
	" -f        set flash flag",
#endif
#endif
	0,
};

static char flashflag;

int
Set(int argc,char *argv[])
{
	char	buf[ENV_ALLOC_SIZE];
	int		opt, i, rc, clear;

	clear = 0;
	rc = CMD_SUCCESS;
	while((opt=getopt(argc,argv,"cf")) != -1) {
		switch(opt) {
		case 'c':		/* clear environment */
			clear = 1;
			break;
#if INCLUDE_FLASHENV
		case 'f':
			flashflag = 1;
			break;
#endif
		default:
			rc = CMD_PARAM_ERROR;
			goto done;
		}
	}

	if (clear) {
		shell_clear();
		goto done;
	}

	if (shell_varcheck(__LINE__) < 0) {
		rc = CMD_FAILURE;
		goto done;
	}

	if (argc == optind) {	/* display all variables */
		shell_print();
	}
	else if (argc == (optind+1)) {	/* clear named var */
		setenv(argv[optind],0);
	}
	else if (argc >= (optind+2)) {	/* Set a specific variable */
		buf[0] = 0;
		for(i=optind+1;i<argc;i++) {
			if ((strlen(buf) + strlen(argv[i]) + 2) >= sizeof(buf)) {
				printf("Too big\n");
				break;
			}
			strcat(buf,argv[i]);
			if (i != (argc-1))
				strcat(buf," ");
		}
		setenv(argv[optind],buf);
	}
	else  {
		rc = CMD_PARAM_ERROR;
	}

done:
	flashflag = 0;
	return(rc);
}

/* Shell variable support routines...
 *	The basic scheme is to malloc in the space needed for the variable
 *	name and the value of that variable.  For each variable that 
 *	exists there is one s_shell structure that is in the linked list.
 *	As shell variables are removed, their corresonding s_shell structure
 *	is NOT removed, but the data pointed to within the structure is
 *	freed.  This keeps the linked list implementation extremely simple
 *	but maintains the versatility gained by using malloc for the 
 *	variables instead of some limted set of static arrays.
 */

/* shell_varcheck():
 * Common point to verify that environment space was successfully
 * allocated...
 */
static int
shell_varcheck(int lno)
{
	if (!shell_vars) {
		printf("Env alloc fail (%d).\n",lno);
		return(-1);
	}
	return(0);
}

/* shell_clear():
 * Clear out the entire environment.
 */
static void
shell_clear(void)
{
	struct	s_shell *sp;

#if INCLUDE_FLASHENV
	if (flashflag) {
		flashenv_erase();
		return;
	}
#endif

	for(sp = shell_vars;sp != (struct s_shell *)0;sp = sp->next) {
		if (sp->name != (char *)0) {
			env_free(sp->name);
			env_free(sp->val);
			sp->name = (char *)0;
			sp->val = (char *)0;
		}
	}
}

/* shell_print():
 *	Print out all of the current shell variables and their values.
 */
static int
shell_print(void)
{
	register struct	s_shell *sp;

#if INCLUDE_FLASHENV
	if (flashflag) {
		flashenv_list(1,0);
		return(0);
	}
#endif

	/* Now that we know the size of the largest variable, we can 
	 * print the list cleanly...
	 */
	sp = shell_vars;
	while(1) {
		if (sp->name != (char *)0) {
			printf("%s: %s\n",sp->name,sp->val);
		}
		if (sp->next != (struct s_shell *)0)
			sp = sp->next;
		else
			break;
	}

	return(0);
}

/* shell_alloc():
 *	First scan through the entire list to see if the requested
 *	shell variable name already exists in the list; if it does,
 *	then just use the same s_shell entry but change the value.
 *	Also, if the new value fits in the same space as the older value,
 *	then just use the same memory space (don't do the free/malloc).
 *	If it doesn't, then scan through the list again.  If there
 *	is one that has no variable assigned to it (name = 0), then
 *	use it for the new allocation.  If all s_shell structures do 
 *	have valid entries, then malloc a new s_shell structure and then
 *	place the new shell variable data in that structure.
 */

static int
shell_alloc(char *name,char *value)
{
	int	namelen, valuelen;
	struct s_shell *sp;

#if INCLUDE_FLASHENV
	if (flashflag) {
		int rc = flashenv_set(name,value);
		flashflag = 0;
		flashenv_list(0,1);
		return(rc);
	}
#endif

	sp = shell_vars;
	namelen = strlen(name);
	valuelen = strlen(value);
	while(1) {
		if (sp->name == (char *)0) {
			if (sp->next != (struct s_shell *)0) {
				sp = sp->next;
				continue;
			}
			else
				break;
		}
		if (strcmp(sp->name,name) == 0) {
			if (sp->vsize < valuelen+1) {		/* If new value is smaller	*/
				env_free(sp->val);				/* than the old value, then */
				sp->val = env_alloc(valuelen+1);/* don't re-allocate any	*/
				if (!sp->val)					/* memory, just copy into	*/
					return(-1);					/* the space used by the	*/
				sp->vsize = valuelen+1;			/* previous value.			*/
			}
			strcpy(sp->val,value);
			return(0);
		}
		if (sp->next == (struct s_shell *)0) 
			break;
		sp = sp->next;
	}
	sp = shell_vars;
	while(1) {
		if (sp->name == (char *)0) {
			sp->name = env_alloc(namelen+1);
			if (!sp->name)
				return(-1);
			strcpy(sp->name,name);
			sp->val = env_alloc(valuelen+1);
			if (!sp->val) {
				env_free(sp->name);
				sp->name = (char *)0;
				return(-1);
			}
			sp->vsize = valuelen+1;
			strcpy(sp->val,value);
			return(0);
		}
		if (sp->next != (struct s_shell *)0)
			sp = sp->next;
		else {
			sp->next = (struct s_shell *)env_alloc(sizeof(struct s_shell));
			if (!sp->next)
				return(-1);
			sp = sp->next;
			sp->name = env_alloc(namelen+1);
			if (!sp->name)
				return(-1);
			strcpy(sp->name,name);
			sp->val = env_alloc(valuelen+1);
			if (!sp->val) {
				env_free(sp->name);
				sp->name = (char *)0;
				return(-1);
			}
			sp->vsize = valuelen+1;
			strcpy(sp->val,value);
			sp->next = (struct s_shell *)0;
			return(0);
		}
	}
}

/* shell_dealloc():
 *	Remove the requested shell variable from the list.  Return 0 if
 *	the variable was removed successfully, otherwise return -1.
 */
static int
shell_dealloc(char *name)
{
	struct	s_shell *sp;

#if INCLUDE_FLASHENV
	if (flashflag) {
		int rc = flashenv_del(name);
		flashflag = 0;
		flashenv_list(0,1);
		return(rc);
	}
#endif

	sp = shell_vars;
	while(1) {
		if (sp->name == (char *)0) {
			if (sp->next == (struct s_shell *)0)
				return(-1);
			else {
				sp = sp->next;
				continue;
			}
		}
		if (strcmp(name,sp->name) == 0) {
			env_free(sp->name);
			env_free(sp->val);
			sp->name = (char *)0;
			sp->val = (char *)0;
			return(0);
		}
		
		if (sp->next == (struct s_shell *)0)
			return(-1);
		else
			sp = sp->next;
	}
}

/* ShellVarInit();
 *	Setup the shell_vars pointer appropriately for additional
 *	shell variable assignments that will be made through shell_alloc().
 */
int
ShellVarInit()
{
	char	buf[16];

#if !INCLUDE_MALLOC
	memset((char *)&envSpace,0,sizeof(envSpace));
#endif

	shell_vars = (struct s_shell *)env_alloc(sizeof(struct s_shell));
	if (shell_varcheck(__LINE__) < 0)
		return(-1);
	
	shell_vars->next = (struct s_shell *)0;
	shell_vars->name = (char *)0;
	sprintf(buf,"0x%lx",APPLICATION_RAMSTART);
	setenv("APPRAMBASE",buf);
	return(0);
}

/* getenv:
 *	Return the pointer to the value entry if the shell variable
 *	name is currently set; otherwise, return a null pointer.
 */
char *
getenv(char *name)
{
	register struct	s_shell *sp;

	for(sp = shell_vars;sp != (struct s_shell *)0;sp = sp->next) {
		if (sp->name != (char *)0) {
			if (strcmp(sp->name,name) == 0)
				return(sp->val);
		}
	}
	return((char *)0);
}

/* setenv:
 *	Interface to shell_dealloc() and shell_alloc().
 */
int
setenv(char *name,char *value)
{
	int rc;

	if (shell_varcheck(__LINE__) < 0)
		return(-1);

printf("setenv(%s,%s)...",name,value);
	if ((value == (char *)0) || (*value == 0))
		rc = shell_dealloc(name);
	else
		rc = shell_alloc(name,value);

printf(rc < 0 ? "failed\n" : "passed\n");
//	if (rc < 0)
//		printf("setenv(%s,%s) failed\n",name,value);
	return(rc);
}


/* shell_sprintf():
 *	Simple way to turn a printf-like formatted string into a shell variable.
 */
int
shell_sprintf(char *varname, char *fmt, ...)
{
	int	tot;
	char buf[CMDLINESIZE];
	va_list argp;

	va_start(argp,fmt);
	tot = vsnprintf(buf,CMDLINESIZE-1,fmt,argp);
	va_end(argp);
	setenv(varname,buf);
	return(tot);
}

